export default class Converter {
  static Dictionary = {
    I: 1,
    V: 5,
    X: 10,
    L: 50,
    C: 100,
    D: 500,
    M: 1000
  };

  static _dictionaryValues = Object.values(this.Dictionary).reverse();
  static _dictionaryNames = Object.keys(this.Dictionary).reverse();

  static _maxDecimalGroupSigns = 3;

  static _decimalToPropagate = null;

  static _validateSign(sign) {
    return sign in this.Dictionary;
  }

  static _isSignDescribesDecimalGroup(sign) {
    return String(this.Dictionary[sign]).substring(0, 1) !== "5";
  }

  static _extractRomanSign(sign, value) {
    const upperDictionaryValue = this.Dictionary[
      this._dictionaryNames[this._dictionaryNames.indexOf(sign) - 1]
    ];

    if (
      value / this.Dictionary[sign] >= 1 &&
      (value / upperDictionaryValue < 1 || upperDictionaryValue === undefined)
    ) {
      console.log("extract", sign, value);
      return value - this.Dictionary[sign];
    } else {
      return false;
    }
  }

  static _extractMultiple(sign, dictionary) {
    const extractResult = this._extractRomanSign(
      sign,
      this._decimalToPropagate
    );

    if (extractResult === 0) {
      dictionary[sign] = dictionary[sign] ? dictionary[sign] + 1 : 1;
      return dictionary;
    }

    if (typeof extractResult === "number") {
      this._decimalToPropagate = extractResult;
      dictionary[sign] = dictionary[sign] ? dictionary[sign] + 1 : 1;

      return this._extractMultiple(sign, dictionary);
    }

    return dictionary;
  }

  static _formatExpression() {}

  static _multiplyString(str, times) {
    let res = "";
    for (let i = 0; i < times; i++) {
      res += str;
    }

    return res;
  }

  static toRoman(decimalValue) {
    let romanExpression = "";
    this._decimalToPropagate = decimalValue;

    const calculatedDictionary = this._dictionaryNames.reduce(
      (dictionary, sign) => this._extractMultiple(sign, dictionary),
      {}
    );

    console.log("calculatedDictionary", calculatedDictionary);

    for (let index = 0; index < this._dictionaryNames.length; index++) {
      const sign = this._dictionaryNames[index];

      if (sign in calculatedDictionary) {
        if (this._isSignDescribesDecimalGroup(sign)) {
          if (calculatedDictionary[sign] > this._maxDecimalGroupSigns) {
            const upperName = this._dictionaryNames[index - 1];
            const result = upperName
              ? `${sign}${upperName}`
              : this._multiplyString(sign, calculatedDictionary[sign]);
            romanExpression += result;
          } else {
            romanExpression += this._multiplyString(
              sign,
              calculatedDictionary[sign]
            );
          }
        } else {
          const lowerName = this._dictionaryNames[index + 1];
          const upperName = this._dictionaryNames[index - 1];
          if (calculatedDictionary[sign] <= 1) {
            if (calculatedDictionary[lowerName] > this._maxDecimalGroupSigns) {
              const result = `${lowerName}${upperName}`;
              romanExpression += result;
            } else {
              romanExpression += `${sign}${this._multiplyString(
                lowerName,
                calculatedDictionary[lowerName]
              )}`;
            }
            calculatedDictionary[lowerName] = 0;
          } else {
            console.error("Cannot be greater than 1");
          }
        }
      }
    }

    return romanExpression;
  }

  // fromRoman logic below

  static _extractSeqSameSigns(signs, targetSign, seqIndex) {
    let result = 0;
    for (let i = 0; i < this._maxDecimalGroupSigns; i++) {
      if (signs[seqIndex + i] === targetSign) {
        result += 1;
      } else {
        break;
      }
    }

    if (result > this._maxDecimalGroupSigns) {
      console.error(
        `Seq row error: should be less than ${this._maxDecimalGroupSigns + 1}`
      );
    }

    return result;
  }

  static fromRoman(romanExpression) {
    let result = 0;
    const signs = romanExpression.split("");

    for (let seqIndex = 0; seqIndex < signs.length; ) {
      const currentSign = signs[seqIndex];
      const upperDictionarySign = this._dictionaryNames[
        this._dictionaryNames.indexOf(currentSign) - 1
      ];

      if (upperDictionarySign === signs[seqIndex + 1]) {
        // records like IX CD CM XL
        const representChunk =
          this.Dictionary[upperDictionarySign] - this.Dictionary[currentSign];
        result += representChunk;
        seqIndex++;
      } else {
        // records like CCC XX III
        const timesToMultiply = this._extractSeqSameSigns(
          signs,
          currentSign,
          seqIndex
        );
        const representChunk = this.Dictionary[currentSign] * timesToMultiply;
        result += representChunk;
        seqIndex += timesToMultiply;
      }
    }

    return result;
  }
}
